# This file is part of Xpra.
# Copyright (C) 2025 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import os.path
from math import ceil
from typing import Any
from time import monotonic
from collections.abc import Callable, Sequence

from xpra.common import noop, ConnectionMessage
from xpra.os_util import gi_import
from xpra.util.objects import typedict
from xpra.util.str_fn import csv
from xpra.net.common import PacketType
from xpra.net.compression import Compressed
from xpra.net.protocol.socket_handler import SocketProtocol
from xpra.codecs.image import ImageWrapper, PlanarFormat
from xpra.codecs.constants import COMPRESS_RATIO, COMPRESS_FMT_SUFFIX
from xpra.gtk.signals import register_os_signals, register_SIGUSR_signals
from xpra.server.base import ServerBase, SERVER_BASES
from xpra.codecs.video import getVideoHelper
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("server", "encoder")
compresslog = Logger("compress")

# ensure we don't create loops!
codec_key = "xpra.codecs.remote"
assert codec_key not in sys.modules
sys.modules[codec_key] = None


SAVE_TO_FILE = os.environ.get("XPRA_SAVE_TO_FILE")
assert (not SAVE_TO_FILE) or SAVE_TO_FILE in ("jpeg", "png", "webp")

COMPRESS_FMT = (
    "compress: %5.1fms for %4ix%-4i pixels at %4i,%-4i               using %9s" + COMPRESS_RATIO + COMPRESS_FMT_SUFFIX
)


def add_device_context(ss, options: dict):
    device_context = ss.allocate_cuda_device_context()
    log(f"add_device_context: cuda_device_context={device_context}")
    if device_context:
        options["cuda-device-context"] = device_context


def _make_video_encoder(encoding: str, src_format: str, codec_type=""):
    specs = getVideoHelper().get_encoder_specs(encoding).get(src_format, ())
    if not specs:
        return None

    def find_spec():
        for espec in specs:
            if codec_type and espec.codec_type != codec_type:
                continue
            assert espec.encoding == encoding
            assert espec.input_colorspace == src_format
            return espec
        return specs[0]

    spec = find_spec()
    return spec.codec_class()


def csc_image(image: ImageWrapper, format_options: Sequence[str]) -> ImageWrapper | None:
    pixel_format = image.get_pixel_format()
    width = image.get_width()
    height = image.get_height()
    for fmt in format_options:
        for csc_spec in getVideoHelper().get_csc_specs(pixel_format).get(fmt, ()):
            log(f"csc {pixel_format!r} -> {csc_spec.codec_type!r} via {csc_spec!r}")
            converter = csc_spec.codec_class()
            converter.init_context(width, height, pixel_format, width, height, fmt, typedict())
            result = converter.convert_image(image)
            converter.clean()
            log(f" -> {result}")
            return result
    return None


class EncoderServer(ServerBase):

    def __init__(self):
        log(f"EncoderServer.__init__() {SERVER_BASES=}")
        super().__init__()
        self.session_type = "encoder"
        self.main_loop = GLib.MainLoop()
        self.encoders: dict[str, dict[int, Any]] = {}

    def __repr__(self):
        return "EncoderServer"

    def init(self, opts) -> None:
        super().init(opts)
        from xpra.codecs.pillow.encoder import get_encodings
        encodings = get_encodings()
        if self.encoding not in ("auto", ) and self.encoding not in encodings:
            raise ValueError(f"unsupported encoding {self.encoding!r}")
        # default to True rather than None (aka "auto"):
        self.sharing = self.sharing is not False

    def install_signal_handlers(self, callback: Callable[[int], None]) -> None:
        sstr = self.get_server_mode() + " server"
        register_os_signals(callback, sstr)
        register_SIGUSR_signals(sstr)

    def do_run(self) -> None:
        log("do_run() calling %s", self.main_loop.run)
        self.main_loop.run()
        log("do_run() end of %()", self.main_loop.run)

    def do_quit(self) -> None:
        log("do_quit: calling main_loop.quit()")
        self.main_loop.quit()
        # from now on, we can't rely on the main loop:
        from xpra.util.system import register_SIGUSR_signals
        register_SIGUSR_signals()

    def cleanup_source(self, source) -> None:
        encoders = self.encoders.pop(source.uuid, {})
        if encoders:
            for seq, encoder in encoders.items():
                try:
                    encoder.clean()
                except RuntimeError:
                    log.error(f"Error cleaning encoder {encoder} for sequence {seq} of connection {source}")
        super().cleanup_source(source)

    def init_packet_handlers(self) -> None:
        super().init_packet_handlers()
        self.add_packets("encode", "context-request", "context-compress", "context-close")

    def parse_hello(self, ss, c: typedict, send_ui: bool) -> None:
        super().parse_hello(ss, c, send_ui)
        from xpra.server.source.encodings import EncodingsMixin
        if not isinstance(ss, EncodingsMixin):
            raise ValueError("client did not enable encoding")

    def add_new_client(self, ss, c: typedict, send_ui: bool, share_count: int) -> None:
        super().add_new_client(ss, c, send_ui, share_count)
        ss.protocol.large_packets.append("encode-response")

    def _process_encode(self, proto: SocketProtocol, packet: PacketType) -> None:
        # this function is only used by the `encode` client,
        # not the `remote` encoder
        ss = self.get_server_source(proto)
        if not ss:
            return
        input_coding, pixel_format, raw_data, width, height, rowstride, options, metadata = packet[1:9]
        depth = 32
        bpp = 4
        full_range = True
        encoding = "png" if ss.encoding in ("auto", "") else ss.encoding
        log("encode request from %s, %s to %s (from %s)", ss, input_coding, encoding, ss.encoding)
        # connection encoding options:
        eo = dict(ss.default_encoding_options)
        # the request can override:
        eo.update(options)
        log("using settings: %s", eo)
        free = noop
        rgb_data = raw_data
        if input_coding == "mmap":
            if not ss.mmap_supported or not ss.mmap_read_area:
                raise RuntimeError("mmap packet but mmap read is not available")
            chunks = options.pop("chunks", ())
            rgb_data, free = ss.mmap_read_area.mmap_read(*chunks)
        if options.get("lz4") > 0:
            from xpra.net.lz4.lz4 import decompress
            rgb_data = decompress(rgb_data, max_size=64*1024*1024)
            free()
            free = noop

        try:
            image = ImageWrapper(0, 0, width, height, rgb_data, pixel_format, depth, rowstride,
                                 bpp, PlanarFormat.PACKED, True, None, full_range)
            if SAVE_TO_FILE:
                from xpra.codecs.debug import save_imagewrapper
                now = int(monotonic()*1000)
                filename = f"{now}.{SAVE_TO_FILE}"
                save_imagewrapper(image, filename)
                log.info(f"saved {image}, pixels={len(rgb_data)} {type(rgb_data)} to {filename!r}")
            from xpra.codecs.pillow.encoder import encode, get_encodings
            if encoding in get_encodings():
                # simple path: use pillow
                coding, compressed, client_options, width, height, stride, bpp = encode(encoding, image, typedict(eo))
                bdata = compressed.data
            else:
                # try a video encoder:
                encoder = _make_video_encoder(encoding, pixel_format)
                if not encoder:
                    msg = f"no video encoders found for {encoding!r} and {pixel_format!r}"
                    log(msg)
                    especs = getVideoHelper().get_encoder_specs(encoding)
                    input_cs_options = ()
                    if especs:
                        input_cs_options = tuple(especs.keys())
                        image = csc_image(image, input_cs_options)
                        if image:
                            pixel_format = image.get_pixel_format()
                            encoder = _make_video_encoder(encoding, pixel_format)
                    if not encoder:
                        log(f" supported pixel formats for {encoding!r}: %s", csv(input_cs_options))
                        raise ValueError(msg)
                add_device_context(ss, options)
                encoder.init_context(encoding, width, height, pixel_format, typedict(options))
                bdata, client_options = encoder.compress_image(image, typedict(options))
                bpp = 24
                stride = 0
                coding = encoding
        except (ValueError, RuntimeError) as e:
            log("encode failed", exc_info=True)
            self.disconnect_client(proto, ConnectionMessage.SERVER_ERROR, f"failed to encode: {e}")
            return
        finally:
            free()
        packet = ["encode-response", coding, bdata, client_options, width, height, stride, bpp, metadata]
        ss.send_async(*packet)

    def _process_context_close(self, proto: SocketProtocol, packet: PacketType) -> None:
        ss = self.get_server_source(proto)
        if not ss:
            return
        seq, message = packet[1:3]
        encoder = self.encoders.get(ss.uuid, {}).pop(seq, None)
        if not encoder:
            log(f"closing: encoder not found for uuid {ss.uuid!r} and sequence {seq}")
            return
        log(f"context-close: {encoder!r}, {message=}")
        try:
            encoder.clean()
        except RuntimeError:
            log.error(f"Error cleaning encoder {encoder} for sequence {seq} of connection {ss}")

    def _process_context_request(self, proto: SocketProtocol, packet: PacketType) -> None:
        ss = self.get_server_source(proto)
        if not ss:
            return
        seq, codec_type, encoding, width, height, src_format, options = packet[1:8]
        try:
            encoder = _make_video_encoder(encoding, src_format, codec_type)
            add_device_context(ss, options)
            encoder.init_context(encoding, width, height, src_format, typedict(options))
            self.encoders.setdefault(ss.uuid, {})[seq] = encoder
            log(f"new encoder: {encoder}")
            ss.send("context-response", seq, True, "", encoder.get_info())
        except RuntimeError as e:
            log("context request failed", exc_info=True)
            ss.send("context-response", seq, False, f"initialization error: {e}", {})

    def _process_context_compress(self, proto: SocketProtocol, packet: PacketType) -> None:
        ss = self.get_server_source(proto)
        if not ss:
            return
        seq, metadata, pixels, options, send_opts = packet[1:6]
        encoder = self.encoders.get(ss.uuid, {}).get(seq)
        if not encoder:
            log.error(f"Error encoder not found for uuid {ss.uuid!r} and sequence {seq}")
            ss.send("context-data", seq, b"", {"error": "context not found"}, {})
            return
        encoding = encoder.get_encoding()
        free_cb = []

        def free_all() -> None:
            for free_fn in free_cb:
                free_fn()
            free_cb[:] = []

        chunks = send_opts.pop("chunks", ())
        planes = metadata.get("planes", 0)
        log("compress request with %i planes, mmap chunks=%s", planes, chunks)
        start = monotonic()
        if not pixels and chunks:
            # get the pixels from the mmap chunks:
            if planes == PlanarFormat.PACKED:
                pixels, free = ss.mmap_read_area.mmap_read(*chunks)
                free_cb.append(free)
            else:
                pixels = []
                for plane in range(planes):
                    plane_pixels, free = ss.mmap_read_area.mmap_read(*chunks[plane])
                    free_cb.append(free)
                    pixels.append(plane_pixels)

        if send_opts.get("lz4", 0) > 0:
            from xpra.net.lz4.lz4 import decompress
            if planes == PlanarFormat.PACKED:
                pixels = decompress(pixels)
            else:
                pixels = [decompress(plane) for plane in pixels]
            free_all()

        metadata["pixels"] = pixels
        try:
            image = ImageWrapper(**metadata)
            log(f"{encoder=} {image=}")
            add_device_context(ss, options)
            bdata, client_options = encoder.compress_image(image, typedict(options))
        finally:
            free_all()

        delayed = client_options.get("delayed", 0)
        if bdata is None and not delayed:
            log.warn(f"Warning: no data from encoder {encoder}")
            log.warn(" options:%s", client_options)
            ss.send("context-data", seq, b"", {"error": "no data from encoder"}, {})
            return
        reply_opts = {}
        data = Compressed(encoding, bdata or b"")
        if len(data) >= 4096:
            mmap_write_area = getattr(ss, "mmap_write_area", None)
            log(f"{len(bdata)} bytes, {client_options=}, {mmap_write_area=}")
            if mmap_write_area:
                reply_opts["chunks"] = mmap_write_area.write_data(bdata)
                data = b""
        ss.send("context-data", seq, data, client_options, reply_opts)
        end = monotonic()
        csize = len(bdata or b"")
        psize = image.get_bytesperpixel() * image.get_width() * image.get_height()
        x = image.get_x()
        y = image.get_y()
        outw = encoder.get_width()
        outh = encoder.get_height()
        compresslog(COMPRESS_FMT,
                    (end-start) * 1000, outw, outh, x, y, encoding,
                    100.0*csize/psize, ceil(psize/1024), ceil(csize/1024),
                    seq, client_options, options)



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
from collections.abc import Callable


def do_init():
    for x in list(sys.argv):
        if x.startswith("-psn_"):
            sys.argv.remove(x)
    if os.environ.get("XPRA_HIDE_DOCK", "") == "1":
        from AppKit import NSApp
        # NSApplicationActivationPolicyAccessory = 1
        NSApp.setActivationPolicy_(1)


def do_init_env():
    from xpra.platform import init_env_common
    init_env_common()
    if os.environ.get("CRYPTOGRAPHY_OPENSSL_NO_LEGACY") is None:
        os.environ["CRYPTOGRAPHY_OPENSSL_NO_LEGACY"] = "1"
    # GStreamer's paths:
    bundle_contents = os.environ.get("GST_BUNDLE_CONTENTS")
    if bundle_contents:
        rsc_dir = os.path.join(bundle_contents, "Resources")
        if "GST_PLUGIN_PATH" not in os.environ:
            os.environ["GST_PLUGIN_PATH"] = os.path.join(rsc_dir, "lib", "gstreamer-1.0")
        if "GST_PLUGIN_SCANNER" not in os.environ:
            os.environ["GST_PLUGIN_SCANNER"] = os.path.join(rsc_dir, "libexec", "gstreamer-1.0", "gst-plugin-scanner")


def default_gtk_main_exit():
    from xpra.os_util import gi_import
    gtk = gi_import("Gtk")
    gtk.main_quit()


exit_cb: Callable = default_gtk_main_exit


def quit_handler(*_args):
    exit_cb()
    return True


def set_exit_cb(ecb: Callable):
    global exit_cb
    assert ecb is not None
    exit_cb = ecb


macapp = None


def get_OSXApplication():
    global macapp
    if macapp is None:
        from xpra.os_util import gi_import
        osxapp = gi_import("GtkosxApplication")
        macapp = osxapp.Application()
        macapp.connect("NSApplicationWillTerminate", quit_handler)
    return macapp


# workaround for Big Sur dylib cache mess:
# https://stackoverflow.com/a/65599706/428751
def patch_find_library():
    from ctypes import util  # pylint: disable=import-outside-toplevel
    orig_util_find_library = util.find_library

    def new_util_find_library(name):
        res = orig_util_find_library(name)
        if res:
            return res
        return '/System/Library/Frameworks/' + name + '.framework/' + name

    util.find_library = new_util_find_library


if os.environ.get("XPRA_OSX_PATCH_FIND_LIBRARY", "1") == "1":
    patch_find_library()



#!/usr/bin/env python3

from cairo import OPERATOR_SOURCE   # pylint: disable=no-name-in-module

import gi
gi.require_version("Gtk", "3.0")  # @UndefinedVariable
from gi.repository import Gtk, GLib  # pylint: disable=wrong-import-position @UnresolvedImport


WIDTH, HEIGHT = 640, 640

class FPSWindow(Gtk.Window):

    def __init__(self):
        super().__init__()
        self.set_default_size(WIDTH, HEIGHT)
        self.set_app_paintable(True)
        self.counter = 0
        self.connect("draw", self._draw)
        self.connect("destroy", Gtk.main_quit)
        GLib.timeout_add(10, self.repaint)


    def on_key_press(self, *args):
        pass

    def repaint(self):
        self.counter += 1
        self.queue_draw()
        return True

    def _draw(self, widget, cr):
        c = 0.2
        def paint_block(x, y, w, h, div):
            split_h = self.counter//div % h
            #top half:
            if split_h>0:
                cr.new_path()
                cr.set_operator(OPERATOR_SOURCE)
                cr.set_source_rgb(c, c, c)
                cr.rectangle(x, y, w, split_h)
                cr.fill()
            #bottom half:
            if split_h<h:
                cr.new_path()
                cr.set_operator(OPERATOR_SOURCE)
                cr.set_source_rgb(0, 0, 0)
                cr.rectangle(x, y+split_h, w, h-split_h)
                cr.fill()
            #show label:
            cr.set_source_rgb(1, 1, 1)
            cr.move_to(x+w/2-12, y+h/2+8)
            cr.show_text("1/%s" % div)

        w, h = widget.get_size()
        paint_block(0, 0, w//2, h//2, 1)
        if self.counter%2==0:        #half-rate
            paint_block(w//2, 0, w//2, h//2, 2)
        if self.counter%4==0:        #quarter rate
            paint_block(0, h//2, w//2, h//2, 4)
        if self.counter%8==0:        #one-eigth rate
            paint_block(w//2, h//2, w//2, h//2, 8)

window = FPSWindow()
window.show_all()
Gtk.main()



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import time
import asyncio
from typing import Any
from collections.abc import Awaitable, Callable

from queue import SimpleQueue
from collections import namedtuple
from collections.abc import Coroutine, Generator

from time import monotonic
from xpra.scripts.config import InitExit
from xpra.util.thread import start_thread
from xpra.util.str_fn import csv
from xpra.util.env import envbool
from xpra.os_util import WIN32
from xpra.log import Logger

log = Logger("quic")

UVLOOP = envbool("XPRA_UVLOOP", not WIN32)

ExceptionWrapper = namedtuple("ExceptionWrapper", ("exception", "args"))


class ThreadedAsyncioLoop:
    """
    shim for quic asyncio sockets,
    this runs the asyncio main loop in a thread
    and provides methods for:
    * calling functions as tasks
    * turning an async function into a sync function
     (for calling async functions from regular threads)
    """

    def __init__(self):
        self.loop: asyncio.AbstractEventLoop | None = None
        start_thread(self.run_forever, "asyncio-thread", True)
        self.wait_for_loop()

    def run_forever(self) -> None:
        if UVLOOP:
            try:
                # noinspection PyPackageRequirements
                import uvloop  # pylint: disable=import-outside-toplevel
            except ImportError:
                log.warn("Warning: uvloop not found")
            else:
                log("installing uvloop")
                uvloop.install()
                log.info(f"uvloop {uvloop.__version__} installed")
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        self.loop = loop
        self.loop.run_forever()
        self.loop.close()

    def wait_for_loop(self) -> None:
        now = monotonic()
        while monotonic() - now < 1 and self.loop is None:
            log("waiting for asyncio event loop")
            time.sleep(0.01)
        if self.loop is None:
            raise RuntimeError("no asyncio main loop")

    def call(self, f: Callable | Coroutine | Generator) -> None:
        log(f"call({f})")
        if not self.loop:
            raise RuntimeError("no main loop")
        if isinstance(f, (Coroutine, Generator)):

            def tsafe() -> None:
                log(f"creating task for {f}")
                assert self.loop
                self.loop.create_task(f)

            self.loop.call_soon_threadsafe(tsafe)
        else:
            self.loop.call_soon_threadsafe(f)

    def sync(self, async_fn: Callable[..., Awaitable[Any]], *args) -> Any:
        response: SimpleQueue[Any] = SimpleQueue()

        async def awaitable():
            log("awaitable()")
            try:
                a = await async_fn(*args)
                response.put(a)
            except InitExit as e:
                log(f"error calling async function {async_fn} with {args}", exc_info=True)
                response.put(ExceptionWrapper(InitExit, (e.status, str(e))))
            except Exception as e:
                log(f"error calling async function {async_fn} with {args}", exc_info=True)
                response.put(ExceptionWrapper(RuntimeError, (str(e),)))

        def tsafe() -> None:
            a = awaitable()
            log(f"awaitable={a}")
            assert self.loop
            f = asyncio.run_coroutine_threadsafe(a, self.loop)
            log(f"run_coroutine_threadsafe({a}, {self.loop})={f}")

        assert self.loop
        self.loop.call_soon_threadsafe(tsafe)
        log("sync: waiting for response")
        r = response.get()
        if isinstance(r, ExceptionWrapper):
            log(f"sync: re-throwing {r}")
            try:
                instance = r.exception(*r.args)
            except Exception:
                log(f"failed to re-throw {r.exception}{r.args}", exc_info=True)
                raise RuntimeError(csv(r.args)) from None
            else:
                raise instance
        log(f"sync: response={r}")
        return r


singleton: ThreadedAsyncioLoop | None = None


def get_threaded_loop() -> ThreadedAsyncioLoop:
    global singleton
    if not singleton:
        singleton = ThreadedAsyncioLoop()
    return singleton



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.
# pylint: disable-msg=E1101

import sys
import logging
import traceback
from time import monotonic
from threading import Lock
from typing import Any
from collections.abc import Callable

from xpra.util.str_fn import repr_ellipsized, memoryview_to_bytes
from xpra.net.common import PacketType
from xpra.scripts.config import FALSE_OPTIONS, TRUE_OPTIONS
from xpra.server.mixins.stub_server_mixin import StubServerMixin
from xpra.log import Logger, set_global_logging_handler

log = Logger("server")


class LoggingServer(StubServerMixin):
    """
    Mixin for servers that can receive and send logging packets
    """
    PREFIX = "logging"

    def __init__(self):
        self.remote_logging_send: bool = False
        self.remote_logging_receive: bool = False
        self.logging_lock: Lock = Lock()
        self.log_both: bool = False
        self.in_remote_logging: bool = False
        self.local_logging: Callable | None = None
        self.logging_clients: dict[Any, float] = {}

    def init(self, opts) -> None:
        self.log_both = (opts.remote_logging or "").lower() == "both"
        if opts.remote_logging.lower() not in FALSE_OPTIONS:
            self.remote_logging_send = opts.remote_logging.lower() in ("allow", "send", "both")
            # "yes" is here for backwards compatibility:
            self.remote_logging_receive = opts.remote_logging.lower() in ["allow", "receive", "both"] + list(
                TRUE_OPTIONS)

    def cleanup(self) -> None:
        self.stop_capturing_logging()

    def get_server_features(self, _source=None) -> dict[str, Any]:
        return {"remote-logging": self._features()}

    def _features(self) -> dict[str, Any]:
        return {
            "receive": self.remote_logging_receive,
            "send": self.remote_logging_send,
        }

    def get_info(self, _proto):
        return {
            LoggingServer.PREFIX: self._features(),
        }

    def cleanup_protocol(self, protocol) -> None:
        if protocol in self.logging_clients:
            del self.logging_clients[protocol]

    def remove_logging_client(self, protocol) -> None:
        if self.logging_clients.pop(protocol, None) is None:
            log.warn("Warning: logging was not enabled for '%r'", protocol)
        if not self.logging_clients:
            self.stop_capturing_logging()

    def add_logging_client(self, protocol) -> None:
        n = len(self.logging_clients)
        if protocol in self.logging_clients:
            log.warn("Warning: logging already enabled for client %s", protocol)
            return
        log.info("sending log output to %s", protocol)
        self.logging_clients[protocol] = monotonic()
        if n == 0:
            self.start_capturing_logging()

    def start_capturing_logging(self) -> None:
        if not self.local_logging:
            self.local_logging = set_global_logging_handler(self.remote_logging_handler)

    def stop_capturing_logging(self) -> None:
        ll = self.local_logging
        if ll:
            self.local_logging = None
            set_global_logging_handler(ll)

    def remote_logging_handler(self, log, level: int, msg, *args, **kwargs) -> None:
        # prevent loops (if our send call ends up firing another logging call):
        if self.in_remote_logging:
            return
        ll = self.local_logging

        def local_warn(*args) -> None:
            if ll:
                ll(log, logging.WARNING, *args)

        def local_err(message, e=None) -> None:
            if self._closing:
                return
            local_warn("Warning: %s:", message)
            if e:
                local_warn(" %s", e)
            local_warn(" original unformatted message: %s", msg)
            if args:
                local_warn(" %i arguments: %s", len(args), args)
            else:
                local_warn(" (no arguments)")
            try:
                if ll:
                    ll(log, level, msg, *args, **kwargs)
            except Exception:
                pass
            try:
                exc_info = sys.exc_info()
                for x in traceback.format_tb(exc_info[2]):
                    for v in x.splitlines():
                        local_warn(v)
            except Exception:
                pass

        self.in_remote_logging = True
        try:
            if not kwargs.pop("remote", True):
                if ll:
                    try:
                        ll(log, level, msg, *args, **kwargs)
                    except Exception as e:
                        local_warn("Warning: failed to log message locally")
                        local_warn(" %s", e)
                        local_warn(" %s", msg)
                return
            try:
                if args:
                    data = msg % args
                else:
                    data = msg
            except Exception as e:
                local_err("failed to format log message", e)
                return
            for proto, start_time in self.logging_clients.items():
                source = self.get_server_source(proto)
                if not source:
                    continue
                try:
                    dtime = int(1000 * (monotonic() - start_time))
                    if len(data) >= 32:
                        try:
                            data = source.compressed_wrapper("text", data.encode("utf8"), level=1)
                        except Exception:
                            pass
                    source.send("logging", level, data, dtime)
                    exc_info = kwargs.get("exc_info")
                    if exc_info is True:
                        exc_info = sys.exc_info()
                    if exc_info and exc_info[0]:
                        for x in traceback.format_tb(exc_info[2]):
                            self.send("logging", level, x, dtime)
                        try:
                            etypeinfo = exc_info[0].__name__
                        except AttributeError:
                            etypeinfo = str(exc_info[0])
                        source.send("logging", level, "%s: %s" % (etypeinfo, exc_info[1]), dtime)
                except Exception as e:
                    if self._closing:
                        return
                    local_warn("Warning: failed to send log message to %s: %s", source, e)
            if self.log_both and ll:
                try:
                    ll(log, level, msg, *args, **kwargs)
                except Exception:
                    pass
        finally:
            self.in_remote_logging = False

    def _process_logging_control(self, proto, packet: PacketType) -> None:
        action = str(packet[1])
        if action == "start":
            self.add_logging_client(proto)
        elif action == "stop":
            self.remove_logging_client(proto)
        else:
            log.warn("Warning: unknown logging-control action '%r'", action)

    def _process_logging(self, proto, packet: PacketType) -> None:
        self._process_logging_event(proto, packet)

    def _process_logging_event(self, proto, packet: PacketType) -> None:
        assert self.remote_logging_receive
        ss = self.get_server_source(proto)
        if ss is None:
            return
        level, msg = packet[1:3]
        prefix = "client "
        counter = getattr(ss, "counter", 0)
        if counter > 0:
            prefix += "%3i " % counter
        if len(packet) >= 4:
            dtime = packet[3]
            prefix += "@%02i.%03i " % ((dtime // 1000) % 60, dtime % 1000)

        def decode(v) -> str:
            if isinstance(v, str):
                return v
            return memoryview_to_bytes(v).decode("utf8")

        try:
            if isinstance(msg, (tuple, list)):
                dmsg = " ".join(decode(x) for x in msg)
            else:
                dmsg = decode(msg)
            for line in dmsg.splitlines():
                self.do_log(level, prefix + line)
        except Exception as e:
            log("log message decoding error", exc_info=True)
            log.error("Error: failed to parse logging message:")
            log.error(" %s", repr_ellipsized(msg))
            log.estr(e)

    def do_log(self, level, line) -> None:
        with self.logging_lock:
            log.log(level, line)

    def init_packet_handlers(self) -> None:
        if self.remote_logging_receive:
            self.add_packets("logging-event")
            self.add_legacy_alias("logging", "logging-event")
        if self.remote_logging_send:
            self.add_packets("logging-control")



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import socket
from typing import Any

from xpra.platform import platform_import


def get_interface_info(*_args) -> dict[str, Any]:
    return {}


def get_tcp_info(_sock) -> dict[str, Any]:  # pylint: disable=unused-argument
    return {}


platform_import(globals(), "netdev_query", False,
                "get_tcp_info",
                "get_interface_info",
                )


def print_address(iface, addr, defs) -> None:
    from xpra.os_util import POSIX
    from xpra.util.str_fn import print_nested_dict
    for d in defs:
        ip = d.get("addr")
        if ip:
            stype = {
                socket.AF_INET: "IPv4",
                socket.AF_INET6: "IPv6",
            }[addr]
            print(f" * {stype}:     {ip}")
            if POSIX:
                from xpra.net.socket_util import create_tcp_socket
                sock = None
                try:
                    sock = create_tcp_socket(ip, 0)
                    sockfd = sock.fileno()
                    info = get_interface_info(sockfd, iface)
                    if info:
                        print_nested_dict(info, prefix="    ", lchar="-")
                finally:
                    if sock:
                        sock.close()


def print_iface(iface):
    from xpra.os_util import POSIX
    from xpra.net.net_util import import_netifaces
    netifaces = import_netifaces()
    addresses = netifaces.ifaddresses(iface)  # @UndefinedVariable pylint: disable=no-member
    for addr, defs in addresses.items():
        if addr in (socket.AF_INET, socket.AF_INET6):
            print_address(iface, addr, defs)
    if not POSIX:
        info = get_interface_info(0, iface)
        if info:
            print(f"  {info}")


def main():
    # pylint: disable=import-outside-toplevel
    import sys
    from xpra.net.net_util import get_interfaces
    from socket import if_nametoindex
    from xpra.platform import program_context
    from xpra.log import enable_color, consume_verbose_argv
    with program_context("Network-Device-Info", "Network Device Info"):
        enable_color()
        consume_verbose_argv(sys.argv, "network")
        print("Network interfaces found:")
        for iface in get_interfaces():
            try:
                print("* %s (index=%s)" % (iface.ljust(20), if_nametoindex(iface)))
            except OSError:
                print(f"* {iface}")
            print_iface(iface)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import unittest

from unit.server_test_util import ServerTestUtil
from xpra.os_util import POSIX, OSX
from xpra.util.env import OSEnvContext


class TestDisplayUtil(ServerTestUtil):

    def test_display(self):
        from xpra.scripts.server import verify_gdk_display
        with OSEnvContext():
            os.environ["GDK_BACKEND"] = "x11"
            os.environ.pop("DISPLAY", None)
            for d in ("NOTADISPLAY", ""):
                if verify_gdk_display(d):
                    raise RuntimeError(f"{d!r} is not a valid display")

            display = self.find_free_display()
            xvfb = self.start_Xvfb(display)
            os.environ["DISPLAY"] = display
            from xpra.x11.bindings.posix_display_source import X11DisplayContext    #@UnresolvedImport
            with X11DisplayContext(display):
                verify_gdk_display(display)
            xvfb.terminate()


def main():
    #can only work with an X11 server
    if POSIX and not OSX:
        unittest.main()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2021 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
import signal
from subprocess import Popen, PIPE
from collections.abc import Callable

from xpra.common import noerr
from xpra.util.env import envbool
from xpra.os_util import WIN32, OSX, POSIX
from xpra.util.io import use_gui_prompt, which
from xpra.util.system import is_gnome, is_kde
from xpra.util.str_fn import bytestostr
from xpra.scripts.config import FALSE_OPTIONS, TRUE_OPTIONS, InitExit
from xpra.exit_codes import ExitCode
from xpra.log import Logger

log = Logger("exec", "auth")

SKIP_UI: bool = envbool("XPRA_SKIP_UI", False)
PINENTRY: bool = envbool("XPRA_SSH_PINENTRY", POSIX and not OSX)


# pylint: disable=import-outside-toplevel


def get_pinentry_command(setting: str = "yes") -> str:
    log(f"get_pinentry_command({setting})")
    if setting.lower() in FALSE_OPTIONS:
        return ""

    def find_pinentry_bin() -> str:
        if is_gnome():
            return which("pinentry-gnome3")
        if is_kde():
            return which("pinentry-qt")
        return ""

    if setting.lower() in TRUE_OPTIONS:
        return find_pinentry_bin() or which("pinentry")
    if setting == "" or setting.lower() == "auto":
        # figure out if we should use it:
        if WIN32 or OSX:
            # not enabled by default on those platforms
            return ""
        return find_pinentry_bin()
    return setting


def popen_pinentry(pinentry_cmd: str):
    try:
        cmd = [pinentry_cmd]
        if log.is_debug_enabled():
            cmd.append("--debug")
        return Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    except OSError as e:
        log(f"popen_pinentry({pinentry_cmd}) failed", exc_info=True)
        log.error(f"Error: failed to run {pinentry_cmd!r}")
        log.estr(e)
        return None


def run_pinentry(extra_args) -> int:
    messages = list(extra_args)

    def get_input():
        if not messages:
            return None
        return messages.pop(0)

    def process_output(message: str, line: bytes):
        if line.startswith(b"ERR "):
            log.error(f"Error: pinentry responded to {message!r} with:")
            log.error(" %s", line.rstrip(b"\n\r").decode())
        else:
            log(f"pinentry sent {line!r}")

    pinentry_cmd = get_pinentry_command() or "pinentry"
    proc = popen_pinentry(pinentry_cmd)
    if not proc:
        raise InitExit(ExitCode.UNSUPPORTED, "cannot run pinentry")
    return do_run_pinentry(proc, get_input, process_output)


def do_run_pinentry(proc, get_input: Callable, process_output: Callable) -> int:
    message = "connection"
    while proc.poll() is None:
        try:
            line = proc.stdout.readline()
            while process_output(message, line):
                "process_output should eventually return False or None"
            message = get_input()
            if message is None:
                break
            log(f"sending {message!r}")
            r = proc.stdin.write(f"{message}\n".encode())
            proc.stdin.flush()
            log(f"write returned: {r}")
        except OSError:
            log("error running pinentry", exc_info=True)
            break
    if proc.poll() is None:
        proc.terminate()
    noerr(proc.stdin.close)
    noerr(proc.stdout.close)
    noerr(proc.stderr.close)
    exitcode = proc.wait(1)
    if exitcode is None:
        log.warn("Warning: pinentry is still running")
        return -1
    log(f"pinentry ended: {proc.poll()}")
    return exitcode


def pinentry_getpin(pinentry_proc, title: str, description: str, pin_cb: Callable, err_cb: Callable) -> None:
    from urllib.parse import quote
    messages = [
        f"SETPROMPT {quote(title)}",
        f"SETDESC {quote(description)}:",
        "GETPIN",
    ]

    def get_input():
        if not messages:
            return None
        return messages.pop(0)

    def process_output(message: str, output: bytes):
        # log(f"process_output({message}, {output})")
        if message == "GETPIN":
            if output.startswith(b"S "):
                log("getpin message: %s", output[2:].decode())
                # ie: 'S PASSWORD_FROM_CACHE'
                return True  # read more data
            if output.startswith(b"D "):
                pin_value = output[2:].decode().rstrip("\n\r")
                from urllib.parse import unquote
                decoded = unquote(pin_value)
                pin_cb(decoded)
            else:
                err_cb(output.decode().rstrip("\n\r"))
        return False

    do_run_pinentry(pinentry_proc, get_input, process_output)


def run_pinentry_getpin(pinentry_cmd: str, title: str, description: str) -> str:
    proc = popen_pinentry(pinentry_cmd)
    if proc is None:
        return ""
    values: list[str] = []

    def rec(value=""):
        values.append(str(value))

    def err(value=None):
        log("getpin error: %s", value)

    try:
        pinentry_getpin(proc, title, description, rec, err)
    finally:
        noerr(proc.terminate)
    if not values:
        return ""
    return values[0]


def run_pinentry_confirm(pinentry_cmd: str, title: str, prompt: str) -> str:
    proc = popen_pinentry(pinentry_cmd)
    if proc is None:
        return ""
    messages = [
        # we can't use those as the response is multi-line:
        # "GETINFO flavor",
        # "GETINFO version",
        # "GETINFO pid",
    ]
    messages += [
        f"SETPROMPT {title}",
        f"SETDESC {prompt}",
        # "SETKEYINFO %c/%s"
    ]
    messages.append("CONFIRM")
    log("run_pinentry_confirm%s messages=%s", (pinentry_cmd, title, prompt), messages)

    def get_input():
        if not messages:
            return None
        return messages.pop(0)

    confirm_values = []

    def process_output(message, output):
        log("received %s for %s", output, message)
        if message == "CONFIRM":
            confirm_values.append(output.strip(b"\n\r"))

    do_run_pinentry(proc, get_input, process_output)
    if len(confirm_values) != 1:
        return ""
    return bytestostr(confirm_values[0])  # ie: "OK"


def force_focus() -> None:
    from xpra.platform.gui import force_focus as _force_focus
    _force_focus()


def dialog_pass(title: str = "Password Input", prompt: str = "enter password", icon: str = "") -> str:
    log("dialog_pass%s PINENTRY=%s", (title, prompt, icon), PINENTRY)
    if PINENTRY:
        pinentry_cmd = get_pinentry_command()
        if pinentry_cmd:
            return run_pinentry_getpin(pinentry_cmd, title, prompt)

    from xpra.gtk.dialogs.util import dialog_run, do_run_dialog

    def password_input_run() -> str:
        from xpra.gtk.dialogs.pass_dialog import PasswordInputDialogWindow
        dialog = PasswordInputDialogWindow(title, prompt, icon)
        return str(do_run_dialog(dialog))

    return str(dialog_run(password_input_run))


def dialog_confirm(title: str, prompt: str, qinfo=(), icon: str = "", buttons=(("OK", 1),)) -> int:
    from xpra.gtk.dialogs.util import dialog_run, do_run_dialog

    def confirm_run() -> int:
        from xpra.gtk.dialogs.confirm_dialog import ConfirmDialogWindow
        dialog = ConfirmDialogWindow(title, prompt, qinfo, icon, buttons)
        return do_run_dialog(dialog)

    return int(dialog_run(confirm_run))


def confirm(info=(), title: str = "Confirm Key", prompt: str = "Are you sure you want to continue connecting?") -> bool:
    log("confirm%s SKIP_UI=%s, PINENTRY=%s", (info, title, prompt), SKIP_UI, PINENTRY)
    if SKIP_UI:
        return False
    if PINENTRY:
        pinentry_cmd = get_pinentry_command()
        if pinentry_cmd:
            messages = list(info) + ["", prompt]
            return run_pinentry_confirm(pinentry_cmd, title, "%0A".join(messages)) == "OK"
    if use_gui_prompt():
        from xpra.platform.paths import get_icon_filename
        icon = get_icon_filename("authentication", "png") or ""
        NO_CODE = 199
        YES_CODE = 200
        code = dialog_confirm(title, prompt, info, icon, buttons=[("NO", NO_CODE), ("yes", YES_CODE)])
        log("dialog return code=%s", code)
        r = code == YES_CODE
        return r
    log("confirm%s will use stdin prompt", (info, title, prompt))
    prompt = "Are you sure you want to continue connecting (yes/NO)? "
    sys.stderr.write(os.linesep.join(info) + os.linesep + prompt)
    sys.stderr.flush()
    try:
        line = sys.stdin.readline().rstrip(os.linesep)
    except KeyboardInterrupt:
        sys.exit(128 + signal.SIGINT)
    return line.lower() in ("y", "yes")


def input_pass(prompt: str) -> str:
    if SKIP_UI:
        return ""
    if PINENTRY or use_gui_prompt():
        from xpra.platform.paths import get_icon_filename
        icon = get_icon_filename("authentication", "png") or ""
        log(f"input_pass({prompt}) using dialog")
        return dialog_pass("Password Input", prompt, icon)
    from getpass import getpass
    log(f"input_pass({prompt}) using getpass")
    try:
        return getpass(prompt)
    except KeyboardInterrupt:
        sys.exit(128 + signal.SIGINT)


def main() -> int:
    from xpra.platform import program_context
    with program_context("Pinentry-Dialog", "Pinentry Dialog"):
        if "-v" in sys.argv:
            from xpra.log import enable_debug_for
            enable_debug_for("util")

        try:
            return dialog_confirm(*sys.argv[1:])
        except KeyboardInterrupt:
            return 1


if __name__ == "__main__":
    v = main()
    sys.exit(v)



# This file is part of Xpra.
# Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
from typing import Any, TypeAlias
from collections.abc import Callable, Sequence
from time import sleep, monotonic
from threading import Event
from collections import deque
from queue import SimpleQueue

from xpra.util.thread import start_thread
from xpra.common import FULL_INFO, noop
from xpra.util.objects import AtomicInteger, typedict, notypedict
from xpra.util.env import envbool
from xpra.net.common import PacketType, PacketElement
from xpra.net.compression import compressed_wrapper, Compressed, LevelCompressed
from xpra.server.source.source_stats import GlobalPerformanceStatistics
from xpra.server.source.stub_source_mixin import StubSourceMixin
from xpra.log import Logger

log = Logger("server")
notifylog = Logger("notify")

YIELD = envbool("XPRA_YIELD", False)

counter = AtomicInteger()

ENCODE_WORK_ITEM_TUPLE = tuple[bool, Callable, Sequence[Any]]
ENCODE_WORK_ITEM: TypeAlias = ENCODE_WORK_ITEM_TUPLE | None


class ClientConnection(StubSourceMixin):
    """
    This class mediates between the server class
    (which only knows about actual window objects and display server events)
    and the client specific WindowSource instances (which only know about window ids
    and manage window pixel compression).
    It sends messages to the client via its 'protocol' instance (the network connection),
    directly for a number of cases (cursor, audio, notifications, etc.)
    or on behalf of the window sources for pixel data.

    Strategy: if we have 'ordinary_packets' to send, send those.
    When we don't, then send packets from the 'packet_queue'. (compressed pixels or clipboard data)
    See 'next_packet'.

    The UI thread calls damage(), which goes into WindowSource and eventually (batching may be involved)
    adds the damage pixels ready for processing to the encode_work_queue,
    items are picked off by the separate 'encode' thread (see 'encode_loop')
    and added to the damage_packet_queue.
    """

    def __init__(self, protocol, disconnect_cb: Callable, setting_changed: Callable[[str, Any], None]):
        self.counter = counter.increase()
        self.protocol = protocol
        self.connection_time = monotonic()
        self.close_event = Event()
        self.disconnect = disconnect_cb

        # holds actual packets ready for sending (already encoded)
        # these packets are picked off by the "protocol" via 'next_packet()'
        # format: packet, wid, pixels, start_send_cb, end_send_cb
        # (only packet is required - the rest can be 0/None for clipboard packets)
        # the encode work queue is used by mixins that need to encode data before sending it,
        self.packet_queue = deque[tuple[PacketType, int, int, bool]]()
        # ie: encodings and clipboard
        # this queue will hold functions to call to compress data (pixels, clipboard)
        # items placed in this queue are picked off by the "encode" thread,
        # the functions should add the packets they generate to the 'packet_queue'
        self.encode_work_queue: SimpleQueue[None | tuple[bool, Callable, Sequence[Any]]] = SimpleQueue()
        self.encode_thread = None
        self.ordinary_packets: list[tuple[PacketType, bool, bool]] = []

        self.startup_completed = False
        self.suspended = False
        self.client_packet_types = ()
        self.setting_changed = setting_changed
        self.queue_encode: Callable[[ENCODE_WORK_ITEM], None] = self.start_queue_encode

    def run(self) -> None:
        # ready for processing:
        self.protocol.set_packet_source(self.next_packet)

    def __repr__(self) -> str:
        classname = type(self).__name__
        return f"{classname}({self.counter} : {self.protocol})"

    def init_state(self) -> None:
        self.hello_sent = False
        self.share = False
        self.lock = False
        self.client_control_commands: Sequence[str] = ()
        self.xdg_menu = True
        self.ssh_auth_sock = ""
        # what we send back in hello packet:
        self.ui_client = True
        # default 'wants' is not including "events" or "default_cursor":
        self.wants = ["encodings", "versions", "features", "display", "packet-types"]
        # these statistics are shared by all WindowSource instances:
        self.statistics = GlobalPerformanceStatistics()

    def is_closed(self) -> bool:
        return self.close_event.is_set()

    def cleanup(self) -> None:
        log("%s.close()", self)
        self.close_event.set()
        self.protocol = None
        self.statistics.reset(0)

    def suspend(self) -> None:
        self.suspended = True

    def resume(self) -> None:
        self.suspended = False

    def may_notify(self, *args, **kwargs) -> None:
        # fugly workaround,
        # MRO is depth first and would hit the default implementation
        # instead of the mixin unless we force it:
        notification_mixin = sys.modules.get("xpra.server.source.notification")
        if notification_mixin and isinstance(self, notification_mixin.NotificationMixin):
            notification_mixin.NotificationMixin.may_notify(self, *args, **kwargs)

    def compressed_wrapper(self, datatype, data, **kwargs) -> Compressed | LevelCompressed:
        # set compression flags based on self.lz4:
        kw = {"lz4": getattr(self, "lz4", False)}
        kw.update(kwargs)
        return compressed_wrapper(datatype, data, can_inline=False, **kw)

    def may_update_bandwidth_limits(self):
        # this method is only available when the NetworkState mixin is enabled:
        update_bandwidth_limits = getattr(self, "update_bandwidth_limits", noop)
        if update_bandwidth_limits != noop:
            update_bandwidth_limits()

    def parse_client_caps(self, c: typedict) -> None:
        # general features:
        self.share = c.boolget("share")
        self.lock = c.boolget("lock")
        self.client_control_commands = c.strtupleget("control_commands")
        self.xdg_menu = c.boolget("xdg-menu", False)
        self.ssh_auth_sock = c.strget("ssh-auth-sock")

    def startup_complete(self) -> None:
        log("startup_complete()")
        self.startup_completed = True
        self.send("startup-complete")

    # The encode thread loop management:
    #
    def start_queue_encode(self, item: ENCODE_WORK_ITEM) -> None:
        # start the encode work queue:
        # holds functions to call to compress data (pixels, clipboard)
        # items placed in this queue are picked off by the "encode" thread,
        # the functions should add the packets they generate to the 'packet_queue'
        self.queue_encode = self.encode_work_queue.put
        self.queue_encode(item)
        self.encode_thread = start_thread(self.encode_loop, "encode")

    def encode_queue_size(self) -> int:
        return self.encode_work_queue.qsize()

    def call_in_encode_thread(self, optional: bool, fn: Callable, *args) -> None:
        """
            This is used by WindowSource to queue damage processing to be done in the 'encode' thread.
            The 'encode_and_send_cb' will then add the resulting packet to the 'packet_queue' via 'queue_packet'.
        """
        self.statistics.compression_work_qsizes.append((monotonic(), self.encode_queue_size()))
        self.queue_encode((optional, fn, args))

    def queue_packet(self, packet: PacketType, wid=0, pixels=0,
                     wait_for_more=False) -> None:
        """
            Add a new 'draw' packet to the 'packet_queue'.
            Note: this code runs in the non-ui thread
        """
        now = monotonic()
        self.statistics.packet_qsizes.append((now, len(self.packet_queue)))
        if wid > 0:
            self.statistics.damage_packet_qpixels.append(
                (now, wid, sum(x[2] for x in tuple(self.packet_queue) if x[1] == wid))
            )
        self.packet_queue.append((packet, wid, pixels, wait_for_more))
        p = self.protocol
        if p:
            p.source_has_more()

    def encode_loop(self) -> None:
        """
            This runs in a separate thread and calls all the function callbacks
            which are added to the 'encode_work_queue'.
            Must run until we hit the end of queue marker,
            to ensure all the queued items get called,
            those that are marked as optional will be skipped when is_closed()
        """
        while True:
            item = self.encode_work_queue.get(True)
            if item is None:
                return  # empty marker
            # some function calls are optional and can be skipped when closing:
            # (but some are not, like encoder clean functions)
            optional_when_closing, fn, args = item
            if optional_when_closing and self.is_closed():
                continue
            try:
                fn(*args)
            except Exception as e:
                if self.is_closed():
                    log("ignoring encoding error calling %s because the source is already closed:", item)
                    log(" %s", e)
                else:
                    log.error("Error during encoding:", exc_info=True)
                del e
            if YIELD:
                sleep(0)

    ######################################################################
    # network:
    def next_packet(self) -> tuple[PacketType, bool, bool]:
        """ Called by protocol.py when it is ready to send the next packet """
        if self.is_closed():
            return ("closed", ), False, False
        synchronous = True
        more = False
        if self.ordinary_packets:
            packet, synchronous, more = self.ordinary_packets.pop(0)
        elif self.packet_queue:
            packet, _, _, more = self.packet_queue.popleft()
        else:
            packet = ("none", )
        if not more:
            more = bool(packet) and bool(self.ordinary_packets or self.packet_queue)
        return packet, synchronous, more

    def send(self, packet_type: str, *parts: PacketElement, **kwargs) -> None:
        """ This method queues non-damage packets (higher priority) """
        synchronous = bool(kwargs.get("synchronous", True))
        will_have_more = bool(kwargs.get("will_have_more", not synchronous))
        p = self.protocol
        if p:
            packet = (packet_type, *parts)
            self.ordinary_packets.append((packet, synchronous, will_have_more))
            p.source_has_more()

    def send_more(self, packet_type: str, *parts: PacketElement, **kwargs) -> None:
        kwargs["will_have_more"] = True
        self.send(packet_type, *parts, **kwargs)

    def send_async(self, packet_type: str, *parts: PacketElement, **kwargs) -> None:
        kwargs["synchronous"] = False
        kwargs["will_have_more"] = False
        self.send(packet_type, *parts, **kwargs)

    ######################################################################
    # info:
    def get_info(self) -> dict[str, Any]:
        if not FULL_INFO:
            return {"protocol": "xpra"}
        info = {
            "protocol": "xpra",
            "connection_time": int(self.connection_time),
            "elapsed_time": int(monotonic() - self.connection_time),
            "counter": self.counter,
            "hello-sent": self.hello_sent,
            "ssh-auth-sock": self.ssh_auth_sock,
            "packet-types": self.client_packet_types,
        }
        p = self.protocol
        if p:
            info["connection"] = p.get_info()
        info.update(self.get_features_info())
        return info

    def get_features_info(self) -> dict[str, Any]:
        info = {
            "lock": bool(self.lock),
            "share": bool(self.share),
            "xdg-menu": bool(self.xdg_menu),
        }
        return info

    def send_info_response(self, info: dict) -> None:
        self.send_async("info-response", notypedict(info))

    def send_setting_change(self, setting: str, value: PacketElement) -> None:
        self.send_more("setting-change", setting, value)

    def send_server_event(self, event_type: str, *args: PacketElement) -> None:
        if "events" in self.wants:
            self.send_more("server-event", event_type, *args)

    def send_client_command(self, command: str, *args: PacketElement) -> None:
        if self.hello_sent:
            self.send_more("control", command, *args)
